
#include <windows.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "bmp.h"
#include <vector>

#include "image.h"

#pragma warning(disable:4996)

static bool readfile(char const *path, int maxlen, std::vector<unsigned char> *out)
{
	bool success = false;
	int fd = open(path, O_RDONLY | O_BINARY);
	if (fd != -1) {
		struct stat st;
		if (fstat(fd, &st) == 0) {
			if (maxlen < 0 || st.st_size < maxlen) {
				success = true;
				int len = st.st_size;
				out->resize(len);
				int pos = 0;
				while (pos < len) {
					int n = len - pos;
					if (n > 65536) {
						n = 65536;
					}
					if (read(fd, &out->at(pos), n) != n) {
						success = false;
						break;
					}
					pos += n;
				}
			}
		}
		close(fd);
	}
	return success;
}

bool read_bmp(image::Image24RGB *img, char const *path)
{
	std::vector<unsigned char> bmp;
	if (!readfile(path, 20000000, &bmp)) {
		return false;
	}
	if (bmp.size() < sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER)) {
		return false;
	}
	BITMAPFILEHEADER *bfh = (BITMAPFILEHEADER *)&bmp[0];
	BITMAPINFOHEADER *bmi = (BITMAPINFOHEADER *)&bmp[sizeof(BITMAPFILEHEADER)];
	unsigned char *data = (unsigned char *)bmi + sizeof(BITMAPINFOHEADER);
	if (memcmp(bfh, "BM", 2) != 0) {
		return false;
	}
	if (bmi->biSize != sizeof(BITMAPINFOHEADER)) {
		return false;
	}

	int width = bmi->biWidth;
	int height = bmi->biHeight;

	if (bmi->biCompression == BI_RGB) {
		if (bmi->biBitCount == 32) {
			img->create(width, height);
			int span = width * 4;
			span -= span & 3;
			if (&bmp[0] + bmp.size() < data + span * height) {
				return false;
			}
			for (int y = 0; y < height; y++) {
				unsigned char const *src = data + span * (height - y - 1);
				image::Pixel24RGB *dst = img->line(y);
				for (int x = 0; x < width; x++) {
					dst->r = src[2];
					dst->g = src[1];
					dst->b = src[0];
					src += 4;
					dst++;
				}
			}
		} else if (bmi->biBitCount == 24) {
			img->create(width, height);
			int span = width * 3 + 3;
			span -= span & 3;
			if (&bmp[0] + bmp.size() < data + span * height) {
				return false;
			}
			for (int y = 0; y < height; y++) {
				unsigned char const *src = data + span * (height - y - 1);
				image::Pixel24RGB *dst = img->line(y);
				for (int x = 0; x < width; x++) {
					dst->r = src[2];
					dst->g = src[1];
					dst->b = src[0];
					src += 3;
					dst++;
				}
			}
		} else if (bmi->biBitCount == 16) {
			img->create(width, height);
			int span = width * 2 + 3;
			span -= span & 3;
			if (&bmp[0] + bmp.size() < data + span * height) {
				return false;
			}
			for (int y = 0; y < height; y++) {
				unsigned char const *src = data + span * (height - y - 1);
				image::Pixel24RGB *dst = img->line(y);
				for (int x = 0; x < width; x++) {
					unsigned short t = (src[1] << 8) | src[0];
					dst->r = ((t >> 10) & 0x1f) * 255 / 31;
					dst->g = ((t >> 5) & 0x1f) * 255 / 31;
					dst->b = (t & 0x1f) * 255 / 31;
					src += 2;
					dst++;
				}
			}
		} else if (bmi->biBitCount == 8) {
			img->create(width, height);
			RGBQUAD *rgb = (RGBQUAD *)data;
			data += sizeof(RGBQUAD) * (bmi->biClrUsed == 0 ? 256 : bmi->biClrUsed);
			int span = width + 3;
			span -= span & 3;
			if (&bmp[0] + bmp.size() < data + span * height) {
				return false;
			}
			for (int y = 0; y < height; y++) {
				unsigned char const *src = data + span * (height - y - 1);
				image::Pixel24RGB *dst = img->line(y);
				if (bmi->biClrUsed == 0) {
					for (int x = 0; x < width; x++) {
						dst->r = dst->g = dst->b = *src;
						src++;
						dst++;
					}
				} else {
					for (int x = 0; x < width; x++) {
						int i = *src;
						if (i < (int)bmi->biClrUsed) {
							dst->r = rgb[i].rgbRed;
							dst->g = rgb[i].rgbGreen;
							dst->b = rgb[i].rgbBlue;
						} else {
							*dst = image::Pixel24RGB();
						}
						src++;
						dst++;
					}
				}
			}
		} else if (bmi->biBitCount == 4) {
			img->create(width, height);
			RGBQUAD *rgb = (RGBQUAD *)data;
			data += sizeof(RGBQUAD) * (bmi->biClrUsed == 0 ? 16 : bmi->biClrUsed);
			int span = (width + 1) / 2 + 3;
			span -= span & 3;
			if (&bmp[0] + bmp.size() < data + span * height) {
				return false;
			}
			for (int y = 0; y < height; y++) {
				unsigned char const *src = data + span * (height - y - 1);
				image::Pixel24RGB *dst = img->line(y);
				unsigned char c = 0;
				for (int x = 0; x < width; x++) {
					if ((x & 1) == 0) {
						c = *src++;
					}
					int i = (c >> 4) & 0x0f;
					c <<= 4;
					dst->r = rgb[i].rgbRed;
					dst->g = rgb[i].rgbGreen;
					dst->b = rgb[i].rgbBlue;
					dst++;
				}
			}
		} else if (bmi->biBitCount == 1) {
			img->create(width, height);
			RGBQUAD *rgb = (RGBQUAD *)data;
			data += sizeof(RGBQUAD) * (bmi->biClrUsed == 0 ? 2 : bmi->biClrUsed);
			int span = (width + 7) / 8 + 3;
			span -= span & 3;
			if (&bmp[0] + bmp.size() < data + span * height) {
				return false;
			}
			for (int y = 0; y < height; y++) {
				unsigned char const *src = data + span * (height - y - 1);
				image::Pixel24RGB *dst = img->line(y);
				unsigned char c = 0;
				for (int x = 0; x < width; x++) {
					if ((x & 7) == 0) {
						c = *src++;
					}
					int i = (c >> 7) & 1;
					c <<= 1;
					dst->r = rgb[i].rgbRed;
					dst->g = rgb[i].rgbGreen;
					dst->b = rgb[i].rgbBlue;
					dst++;
				}
			}
		}
	}

	return true;
}

