
#include "image/image.h"
#include <math.h>
#include <vector>
#include <string.h>

namespace soramimi {
	namespace image {

		//

		static PixelRGB get_pixel(PixelRGB const *src, int w, int h, double x, double y)
		{
			if (w < 1 || h < 1) {
				return PixelRGB();
			}
			x -= 0.5;
			y -= 0.5;
			if (x > w - 1) x = w - 1;
			if (y > h - 1) y = h - 1;
			if (x < 0) x = 0;
			if (y < 0) y = 0;
			int ix = (int)floor(x * 128);
			int iy = (int)floor(y * 128);
			unsigned int rx0 = ix & 127;
			unsigned int ry0 = iy & 127;
			unsigned int rx1 = 128 - rx0;
			unsigned int ry1 = 128 - ry0;
			ix >>= 7;
			iy >>= 7;

			if (rx0 > 0 && ry0 > 0 && ix + 1 < w && iy + 1 < h) {
				PixelRGB p0 = src[w * iy + ix];
				PixelRGB p1 = src[w * iy + (ix + 1)];
				PixelRGB p2 = src[w * (iy + 1) + ix];
				PixelRGB p3 = src[w * (iy + 1) + (ix + 1)];
				if (p0.a + p1.a + p2.a + p3.a > 0) {
					unsigned int t0 = rx1 * ry1;
					unsigned int t1 = rx0 * ry1;
					unsigned int t2 = rx1 * ry0;
					unsigned int t3 = rx0 * ry0;
					unsigned int tt = t0 + t1 + t2 + t3;
					t0 *= p0.a;
					t1 *= p1.a;
					t2 *= p2.a;
					t3 *= p3.a;
					unsigned int a = t0 + t1 + t2 + t3;
					if (a > 0) {
						p0.r = (p0.r * t0 + p1.r * t1 + p2.r * t2 + p3.r * t3) / a;
						p0.g = (p0.g * t0 + p1.g * t1 + p2.g * t2 + p3.g * t3) / a;
						p0.b = (p0.b * t0 + p1.b * t1 + p2.b * t2 + p3.b * t3) / a;
						p0.a = a / tt;
					}
				}
				return p0;
			}

			{
				PixelRGB p0 = src[w * iy + ix];
				if (rx0 > 0 && ix + 1 < w) {
					PixelRGB p1 = src[w * iy + (ix + 1)];
					if (p0.a + p1.a > 0) {
						unsigned int t0 = p0.a * rx1;
						unsigned int t1 = p1.a * rx0;
						unsigned int a = t0 + t1;
						p0.r = (p0.r * t0 + p1.r * t1) / a;
						p0.g = (p0.g * t0 + p1.g * t1) / a;
						p0.b = (p0.b * t0 + p1.b * t1) / a;
						p0.a = a / (rx1 + rx0);
					}
				}
				if (ry0 > 0 && iy + 1 < h) {
					PixelRGB p2 = src[w * (iy + 1) + ix];
					if (rx0 > 0 && ix + 1 < w) {
						PixelRGB p3 = src[w * (iy + 1) + (ix + 1)];
						if (p2.a + p3.a > 0) {
							unsigned int t0 = p2.a * rx1;
							unsigned int t1 = p3.a * rx0;
							unsigned int a = t0 + t1;
							p2.r = (p2.r * t0 + p3.r * t1) / a;
							p2.g = (p2.g * t0 + p3.g * t1) / a;
							p2.b = (p2.b * t0 + p3.b * t1) / a;
							p2.a = a / (rx1 + rx0);
						}
					}
					if (p0.a + p2.a > 0) {
						unsigned int t0 = p0.a * ry1;
						unsigned int t1 = p2.a * ry0;
						unsigned int a = t0 + t1;
						p0.r = (p0.r * t0 + p2.r * t1) / a;
						p0.g = (p0.g * t0 + p2.g * t1) / a;
						p0.b = (p0.b * t0 + p2.b * t1) / a;
						p0.a = a / (ry1 + ry0);
					}
				}
				return p0;
			}
		}

		static PixelY get_pixel(PixelY const *src, int w, int h, double x, double y)
		{
			if (w < 1 || h < 1) {
				return PixelY();
			}
			x -= 0.5;
			y -= 0.5;
			if (x > w - 1) x = w - 1;
			if (y > h - 1) y = h - 1;
			if (x < 0) x = 0;
			if (y < 0) y = 0;
			int ix = (int)floor(x * 128);
			int iy = (int)floor(y * 128);
			unsigned int rx0 = ix & 127;
			unsigned int ry0 = iy & 127;
			unsigned int rx1 = 128 - rx0;
			unsigned int ry1 = 128 - ry0;
			ix >>= 7;
			iy >>= 7;

			if (rx0 > 0 && ry0 > 0 && ix + 1 < w && iy + 1 < h) {
				PixelY p0 = src[w * iy + ix];
				PixelY p1 = src[w * iy + (ix + 1)];
				PixelY p2 = src[w * (iy + 1) + ix];
				PixelY p3 = src[w * (iy + 1) + (ix + 1)];
				if (p0.a + p1.a + p2.a + p3.a > 0) {
					unsigned int t0 = rx1 * ry1;
					unsigned int t1 = rx0 * ry1;
					unsigned int t2 = rx1 * ry0;
					unsigned int t3 = rx0 * ry0;
					unsigned int tt = t0 + t1 + t2 + t3;
					t0 *= p0.a;
					t1 *= p1.a;
					t2 *= p2.a;
					t3 *= p3.a;
					unsigned int a = t0 + t1 + t2 + t3;
					if (a > 0) {
						p0.y = (p0.y * t0 + p1.y * t1 + p2.y * t2 + p3.y * t3) / a;
						p0.a = a / tt;
					}
				}
				return p0;
			}

			{
				PixelY p0 = src[w * iy + ix];
				if (rx0 > 0 && ix + 1 < w) {
					PixelY p1 = src[w * iy + (ix + 1)];
					if (p0.a + p1.a > 0) {
						unsigned int t0 = p0.a * rx1;
						unsigned int t1 = p1.a * rx0;
						unsigned int a = t0 + t1;
						p0.y = (p0.y * t0 + p1.y * t1) / a;
						p0.a = a / (rx1 + rx0);
					}
				}
				if (ry0 > 0 && iy + 1 < h) {
					PixelY p2 = src[w * (iy + 1) + ix];
					if (rx0 > 0 && ix + 1 < w) {
						PixelY p3 = src[w * (iy + 1) + (ix + 1)];
						if (p2.a + p3.a > 0) {
							unsigned int t0 = p2.a * rx1;
							unsigned int t1 = p3.a * rx0;
							unsigned int a = t0 + t1;
							p2.y = (p2.y * t0 + p3.y * t1) / a;
							p2.a = a / (rx1 + rx0);
						}
					}
					if (p0.a + p2.a > 0) {
						unsigned int t0 = p0.a * ry1;
						unsigned int t1 = p2.a * ry0;
						unsigned int a = t0 + t1;
						p0.y = (p0.y * t0 + p2.y * t1) / a;
						p0.a = a / (ry1 + ry0);
					}
				}
				return p0;
			}
		}

		//

		void rop_copy_rgba(PixelRGB *dst, PixelRGB const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				*dst = *src;
				src++;
				dst++;
			}
		}

		void rop_xor_rgba(PixelRGB *dst, PixelRGB const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				if (dst->a != 0 && src->a != 0) {
					dst->r ^= src->r;
					dst->g ^= src->g;
					dst->b ^= src->b;
				}
				src++;
				dst++;
			}
		}

		void rop_min_rgba(PixelRGB *dst, PixelRGB const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				if (dst->a != 0 && src->a != 0) {
					dst->r = std::min(dst->r, src->r);
					dst->g = std::min(dst->g, src->g);
					dst->b = std::min(dst->b, src->b);
				}
				src++;
				dst++;
			}
		}

		void rop_max_rgba(PixelRGB *dst, PixelRGB const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				if (dst->a != 0 && src->a != 0) {
					dst->r = std::max(dst->r, src->r);
					dst->g = std::max(dst->g, src->g);
					dst->b = std::max(dst->b, src->b);
				}
				src++;
				dst++;
			}
		}

		//

		void rop_copy_rgba(PixelY *dst, PixelY const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				*dst = *src;
				src++;
				dst++;
			}
		}

		void rop_xor_rgba(PixelY *dst, PixelY const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				if (dst->a != 0 && src->a != 0) {
					dst->y ^= src->y;
				}
				src++;
				dst++;
			}
		}

		void rop_min_rgba(PixelY *dst, PixelY const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				if (dst->a != 0 && src->a != 0) {
					dst->y = std::min(dst->y, src->y);
				}
				src++;
				dst++;
			}
		}

		void rop_max_rgba(PixelY *dst, PixelY const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				if (dst->a != 0 && src->a != 0) {
					dst->y = std::max(dst->y, src->y);
				}
				src++;
				dst++;
			}
		}

		//

		void rop_copy_rgba(PixelRGB *dst, PixelY const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				*dst = *src;
				src++;
				dst++;
			}
		}

		void rop_xor_rgba(PixelRGB *dst, PixelY const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				if (dst->a != 0 && src->a != 0) {
					dst->r ^= src->y;
					dst->g ^= src->y;
					dst->b ^= src->y;
				}
				src++;
				dst++;
			}
		}

		void rop_min_rgba(PixelRGB *dst, PixelY const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				if (dst->a != 0 && src->a != 0) {
					dst->r = std::min(dst->r, src->y);
					dst->g = std::min(dst->g, src->y);
					dst->b = std::min(dst->b, src->y);
				}
				src++;
				dst++;
			}
		}

		void rop_max_rgba(PixelRGB *dst, PixelY const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				if (dst->a != 0 && src->a != 0) {
					dst->r = std::max(dst->r, src->y);
					dst->g = std::max(dst->g, src->y);
					dst->b = std::max(dst->b, src->y);
				}
				src++;
				dst++;
			}
		}

		//

		void rop_copy_rgba(PixelY *dst, PixelRGB const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				*dst = *src;
				src++;
				dst++;
			}
		}

		void rop_xor_rgba(PixelY *dst, PixelRGB const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				if (dst->a != 0 && src->a != 0) {
					dst->y ^= gray(*src);
				}
				src++;
				dst++;
			}
		}

		void rop_min_rgba(PixelY *dst, PixelRGB const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				if (dst->a != 0 && src->a != 0) {
					dst->y = std::min(dst->y, gray(*src));
				}
				src++;
				dst++;
			}
		}

		void rop_max_rgba(PixelY *dst, PixelRGB const *src, int n, int opacity)
		{
			for (int i = 0; i < n; i++) {
				if (dst->a != 0 && src->a != 0) {
					dst->y = std::max(dst->y, gray(*src));
				}
				src++;
				dst++;
			}
		}

		//

		template <typename D, typename S> void copy_pixel(D *dst, int dst_span, int dst_w, int dst_h, S const *src, int src_span, int src_w, int src_h, int dst_x, int dst_y, int src_x, int src_y, int w, int h, void (*rop)(D *dst, S const *src, int n, int opacity), int opacity)
		{
			if (w < 1) return;
			if (h < 1) return;

			int src_x0 = src_x;
			int src_y0 = src_y;
			int src_x1 = src_x + w;
			int src_y1 = src_y + h;
			int dst_x0 = dst_x;
			int dst_y0 = dst_y;
			int dst_x1 = dst_x + w;
			int dst_y1 = dst_y + h;

			if (dst_x0 < 0)     { src_x0 -= dst_x0;         dst_x0 = 0;     }
			if (src_x0 < 0)     { dst_x0 -= src_x0;         src_x0 = 0;     }
			if (dst_y0 < 0)     { src_y0 -= dst_y0;         dst_y0 = 0;     }
			if (src_y0 < 0)     { dst_y0 -= src_y0;         src_y0 = 0;     }
			if (dst_x1 > dst_w) { src_x1 -= dst_x1 - dst_w; dst_x1 = dst_w; }
			if (src_x1 > src_w) { dst_x1 -= src_x1 - src_w; src_x1 = src_w; }
			if (dst_y1 > dst_h) { src_y1 -= dst_y1 - dst_h; dst_y1 = dst_h; }
			if (src_y1 > src_h) { dst_y1 -= src_y1 - src_h; src_y1 = src_h; }

			if (src_x0 >= src_x1) return;
			if (src_y0 >= src_y1) return;
			if (dst_x0 >= dst_x1) return;
			if (dst_y0 >= dst_y1) return;

			src = src + src_y0 * src_span + src_x0;
			dst = dst + dst_y0 * dst_span + dst_x0;

			w = dst_x1 - dst_x0;
			h = dst_y1 - dst_y0;

			for (int y = 0; y < h; y++) {
				rop(dst, src, w, opacity);
				src += src_span;
				dst += dst_span;
			}
		}

		template <typename T> void fill_pixel(T *dst, int dst_span, int dst_w, int dst_h, T const &src, int x, int y, int w, int h)
		{
			if (w < 1) return;
			if (h < 1) return;

			int dst_x0 = x;
			int dst_y0 = y;
			int dst_x1 = x + w;
			int dst_y1 = y + h;

			if (dst_x0 < 0)     { dst_x0 = 0;     }
			if (dst_y0 < 0)     { dst_y0 = 0;     }
			if (dst_x1 > dst_w) { dst_x1 = dst_w; }
			if (dst_y1 > dst_h) { dst_y1 = dst_h; }

			if (dst_x0 >= dst_x1) return;
			if (dst_y0 >= dst_y1) return;

			dst = dst + dst_y0 * dst_span + dst_x0;

			w = dst_x1 - dst_x0;
			h = dst_y1 - dst_y0;

			for (int y = 0; y < h; y++) {
				for (int x = 0; x < w; x++) {
					dst[x] = src;
				}
				dst += dst_span;
			}
		}

		///////////////////////////////////////////////////////////////////////////////

		void Block::convert(pixel_type_t pt)
		{
			BlockCore *p = _core->convert(pt);
			if (p) {
				assign(p);
			}
		}

		///////////////////////////////////////////////////////////////////////////////

		void BlockCoreRGB::clear()
		{
			PixelRGB *dst = _data[0];
			for (int i = 0; i < 64; i++) {
				for (int j = 0; j < 64; j++) {
					dst->r = 0;
					dst->g = 0;
					dst->b = 0;
					dst->a = 0;
					dst++;
				}
			}
		}

		BlockCore *BlockCoreRGB::clone() const
		{
			BlockCoreRGB *b = new BlockCoreRGB();
			memcpy(b->_data, _data, sizeof(_data));
			return b;
		}

		BlockCore *BlockCoreRGB::convert(pixel_type_t pt) const
		{
			switch (pt) {
			case PT_RGBA:
				return 0;
			case PT_YA:
				{
					BlockCoreY *b = new BlockCoreY();
					PixelRGB const *src = _data[0];
					PixelY *dst = b->_data[0];
					for (int i = 0; i < 64; i++) {
						for (int j = 0; j < 64; j++) {
							dst->y = gray(*src);
							dst->a = src->a;
							src++;
							dst++;
						}
					}
					return b;
				}
			}
			return 0;
		}

		void BlockCoreRGB::fill(int dst_x, int dst_y, int src_w, int src_h, PixelRGB const &src)
		{
			fill_pixel(_data[0], 64, 64, 64, src, dst_x, dst_y, src_w, src_h);
		}

		void BlockCoreRGB::put(int dst_x, int dst_y, PixelRGB const *src, int src_span, int src_x, int src_y, int src_w, int src_h, BlendMode blend, int opacity)
		{
			void (*rop)(PixelRGB *dst, PixelRGB const *src, int n, int opacity);
			rop = 0;
			switch (blend) {
			case BLEND_NORMAL:
				rop = rop_copy_rgba;
				break;
			case BLEND_XOR:
				rop = rop_xor_rgba;
				break;
			case BLEND_MIN:
				rop = rop_min_rgba;
				break;
			case BLEND_MAX:
				rop = rop_max_rgba;
				break;
			}
			if (rop) {
				copy_pixel(_data[0], 64, 64, 64, src, src_span, src_w, src_h, dst_x, dst_y, src_x, src_y, src_w, src_h, rop, opacity);
			}
		}

		void BlockCoreRGB::render(PixelRGB *dst, int dst_span, int dst_w, int dst_h, int dst_x, int dst_y) const
		{
			void (*rop)(PixelRGB *dst, PixelRGB const *src, int n, int opacity);
			rop = rop_copy_rgba;
			copy_pixel(dst, dst_span, dst_w, dst_h, _data[0], 64, 64, 64, dst_x, dst_y, 0, 0, 64, 64, rop, 255);
		}

		void BlockCoreRGB::fill(int dst_x, int dst_y, int src_w, int src_h, PixelY const &src)
		{
			PixelRGB pixel(src.y, src.y, src.y, src.a);
			fill_pixel(_data[0], 64, 64, 64, pixel, dst_x, dst_y, src_w, src_h);
		}

		void BlockCoreRGB::put(int dst_x, int dst_y, PixelY const *src, int src_span, int src_x, int src_y, int src_w, int src_h, BlendMode blend, int opacity)
		{
			void (*rop)(PixelRGB *dst, PixelY const *src, int n, int opacity);
			rop = 0;
			switch (blend) {
			case BLEND_NORMAL:
				rop = rop_copy_rgba;
				break;
			case BLEND_XOR:
				rop = rop_xor_rgba;
				break;
			case BLEND_MIN:
				rop = rop_min_rgba;
				break;
			case BLEND_MAX:
				rop = rop_max_rgba;
				break;
			}
			if (rop) {
				copy_pixel(_data[0], 64, 64, 64, src, src_span, src_w, src_h, dst_x, dst_y, src_x, src_y, src_w, src_h, rop, opacity);
			}
		}

		void BlockCoreRGB::render(PixelY *dst, int dst_span, int dst_w, int dst_h, int dst_x, int dst_y) const
		{
			void (*rop)(PixelY *dst, PixelRGB const *src, int n, int opacity);
			rop = rop_copy_rgba;
			copy_pixel(dst, dst_span, dst_w, dst_h, _data[0], 64, 64, 64, dst_x, dst_y, 0, 0, 64, 64, rop, 255);
		}

		///////////////////////////////////////////////////////////////////////////////

		void BlockCoreY::clear()
		{
			PixelY *dst = _data[0];
			for (int i = 0; i < 64; i++) {
				for (int j = 0; j < 64; j++) {
					dst->y = 0;
					dst->a = 0;
					dst++;
				}
			}
		}

		BlockCore *BlockCoreY::clone() const
		{
			BlockCoreY *b = new BlockCoreY();
			memcpy(b->_data, _data, sizeof(_data));
			return b;
		}

		BlockCore *BlockCoreY::convert(pixel_type_t pt) const
		{
			switch (pt) {
			case PT_YA:
				return 0;
			case PT_RGBA:
				{
					BlockCoreRGB *b = new BlockCoreRGB();
					PixelY const *src = _data[0];
					PixelRGB *dst = b->_data[0];
					for (int i = 0; i < 64; i++) {
						for (int j = 0; j < 64; j++) {
							dst->r = src->y;
							dst->g = src->y;
							dst->b = src->y;
							dst->a = src->a;
							src++;
							dst++;
						}
					}
					return b;
				}
			}
			return 0;
		}

		void BlockCoreY::fill(int dst_x, int dst_y, int src_w, int src_h, PixelRGB const &src)
		{
			PixelY pixel(gray(src), src.a);
			fill_pixel(_data[0], 64, 64, 64, pixel, dst_x, dst_y, src_w, src_h);
		}

		void BlockCoreY::put(int dst_x, int dst_y, PixelRGB const *src, int src_span, int src_x, int src_y, int src_w, int src_h, BlendMode blend, int opacity)
		{
			void (*rop)(PixelY *dst, PixelRGB const *src, int n, int opacity);
			rop = 0;
			switch (blend) {
			case BLEND_NORMAL:
				rop = rop_copy_rgba;
				break;
			case BLEND_XOR:
				rop = rop_xor_rgba;
				break;
			case BLEND_MIN:
				rop = rop_min_rgba;
				break;
			case BLEND_MAX:
				rop = rop_max_rgba;
				break;
			}
			if (rop) {
				copy_pixel(_data[0], 64, 64, 64, src, src_span, src_w, src_h, dst_x, dst_y, src_x, src_y, src_w, src_h, rop, opacity);
			}
		}

		void BlockCoreY::render(PixelRGB *dst, int dst_span, int dst_w, int dst_h, int dst_x, int dst_y) const
		{
			void (*rop)(PixelRGB *dst, PixelY const *src, int n, int opacity);
			rop = rop_copy_rgba;
			copy_pixel(dst, dst_span, dst_w, dst_h, _data[0], 64, 64, 64, dst_x, dst_y, 0, 0, 64, 64, rop, 255);
		}

		void BlockCoreY::fill(int dst_x, int dst_y, int src_w, int src_h, PixelY const &src)
		{
			fill_pixel(_data[0], 64, 64, 64, src, dst_x, dst_y, src_w, src_h);
		}

		void BlockCoreY::put(int dst_x, int dst_y, PixelY const *src, int src_span, int src_x, int src_y, int src_w, int src_h, BlendMode blend, int opacity)
		{
			void (*rop)(PixelY *dst, PixelY const *src, int n, int opacity);
			rop = 0;
			switch (blend) {
			case BLEND_NORMAL:
				rop = rop_copy_rgba;
				break;
			case BLEND_XOR:
				rop = rop_xor_rgba;
				break;
			case BLEND_MIN:
				rop = rop_min_rgba;
				break;
			case BLEND_MAX:
				rop = rop_max_rgba;
				break;
			}
			if (rop) {
				copy_pixel(_data[0], 64, 64, 64, src, src_span, src_w, src_h, dst_x, dst_y, src_x, src_y, src_w, src_h, rop, opacity);
			}
		}

		void BlockCoreY::render(PixelY *dst, int dst_span, int dst_w, int dst_h, int dst_x, int dst_y) const
		{
			void (*rop)(PixelY *dst, PixelY const *src, int n, int opacity);
			rop = rop_copy_rgba;
			copy_pixel(dst, dst_span, dst_w, dst_h, _data[0], 64, 64, 64, dst_x, dst_y, 0, 0, 64, 64, rop, 255);
		}

		///////////////////////////////////////////////////////////////////////////////

		Image::Image()
		: _core(0)
		{
			assign(new ImageCore(PT_RGBA));
		}

		Image::~Image()
		{
			assign(0);
		}

		void Image::assign(ImageCore *p)
		{
			if (p) {
				p->_reference_count++;
			}
			if (_core) {
				if (_core->_reference_count > 1) {
					_core->_reference_count--;
				} else {
					delete _core;
				}
			}
			_core = p;
		}

		void Image::clear(pixel_type_t pt)
		{
			assign(new ImageCore(pt));
		}

		void Image::independent()
		{
			if (_core->_reference_count > 1) {
				Image newimage;
				for (std::map<XY, Block>::iterator it = _core->_blocks.begin(); it != _core->_blocks.end(); it++) {
					_core->_blocks[it->first] = it->second;
				}
			}
		}

		void Image::convert(pixel_type_t pt)
		{
			if (_core->_pixel_type != pt) {
				independent();
				for (std::map<XY, Block>::iterator it = _core->_blocks.begin(); it != _core->_blocks.end(); it++) {
					it->second.convert(pt);
				}
				_core->_pixel_type = pt;
			}
		}

		//template <typename T> Block new_block();
		//template <> Block new_block<PixelRGB>()
		//{
		//	return BlockCoreRGB::NEW();
		//}
		//template <> Block new_block<PixelY>()
		//{
		//	return BlockCoreY::NEW();
		//}

		Block Image::newblock() const
		{
			switch (_core->_pixel_type) {
			case PT_RGBA:
				return BlockCoreRGB::NEW();
			case PT_YA:
				return BlockCoreY::NEW();
			}
			return BlockCoreRGB::NEW();
		}

		template <typename T> void Image::fillT(int x, int y, T const &src, int src_span, int src_x, int src_y, int w, int h)
		{
			int x0 = x;
			int y0 = y;
			int x1 = x + w - 1;
			int y1 = y + h - 1;
			x0 >>= 6;
			y0 >>= 6;
			x1 >>= 6;
			y1 >>= 6;

			int loop_h = y1 - y0 + 1;
			int loop_w = x1 - x0 + 1;

			independent();

			for (int loop_y = 0; loop_y < loop_h; loop_y++) {
				for (int loop_x = 0; loop_x < loop_w; loop_x++) {
					std::map<XY, Block>::iterator it = blocks()->find(XY(x0 + loop_x, y0 + loop_y));
					Block *b;
					if (it == blocks()->end()) {
						it = blocks()->insert(blocks()->end(), std::pair<XY, Block>(XY(x0 + loop_x, y0 + loop_y), newblock()));
						b = &it->second;
					} else {
						b = &it->second;
						b->independent();
					}
					b->_core->fill(x - (x0 + loop_x) * 64, y - (y0 + loop_y) * 64, w, h, src);
				}
			}
		}

		template <typename T> void Image::fillT(int dst_x, int dst_y, int w, int h, T const &src)
		{
			int x, y;

			independent();

			for (y = 0; y < h; y += 64) {
				for (x = 0; x < w; x += 64) {
					int i = w - x;
					int j = h - y;
					if (i > 64) i = 64;
					if (j > 64) j = 64;
					fillT(dst_x + x, dst_y + y, src, i, 0, 0, i, j);
				}
			}
		}

		void Image::fill(int dst_x, int dst_y, int w, int h, PixelRGB const &src)
		{
			fillT(dst_x, dst_y, w, h, src);
		}

		void Image::fill(int dst_x, int dst_y, int w, int h, PixelY const &src)
		{
			fillT(dst_x, dst_y, w, h, src);
		}

		template <typename T> void Image::putT(int x, int y, T *src, int src_span, int src_x, int src_y, int w, int h, BlendMode blend, int opacity)
		{
			int x0 = x;
			int y0 = y;
			int x1 = x + w - 1;
			int y1 = y + h - 1;
			x0 >>= 6;
			y0 >>= 6;
			x1 >>= 6;
			y1 >>= 6;

			int loop_h = y1 - y0 + 1;
			int loop_w = x1 - x0 + 1;

			for (int loop_y = 0; loop_y < loop_h; loop_y++) {
				for (int loop_x = 0; loop_x < loop_w; loop_x++) {
					std::map<XY, Block>::iterator it = blocks()->find(XY(x0 + loop_x, y0 + loop_y));
					Block *b;
					if (it == blocks()->end()) {
						it = blocks()->insert(blocks()->end(), std::pair<XY, Block>(XY(x0 + loop_x, y0 + loop_y), BlockCoreRGB::NEW()));
						b = &it->second;
					} else {
						b = &it->second;
						b->independent();
					}
					b->_core->put(x - (x0 + loop_x) * 64, y - (y0 + loop_y) * 64, src, src_span, src_x, src_y, w, h, blend, opacity);
				}
			}
		}

		void Image::put(int x, int y, PixelRGB *src, int src_span, int src_x, int src_y, int w, int h, BlendMode blend, int opacity)
		{
			putT(x, y, src, src_span, src_x, src_y, w, h, blend, opacity);
		}

		void Image::put(int x, int y, PixelY *src, int src_span, int src_x, int src_y, int w, int h, BlendMode blend, int opacity)
		{
			putT(x, y, src, src_span, src_x, src_y, w, h, blend, opacity);
		}

		template <typename T> void Image::renderT(int x, int y, int w, int h, T *dst) const
		{
			int x0 = x;
			int y0 = y;
			int x1 = x + w - 1;
			int y1 = y + h - 1;
			x0 >>= 6;
			y0 >>= 6;
			x1 >>= 6;
			y1 >>= 6;

			int loop_h = y1 - y0 + 1;
			int loop_w = x1 - x0 + 1;

			for (int loop_y = 0; loop_y < loop_h; loop_y++) {
				for (int loop_x = 0; loop_x < loop_w; loop_x++) {
					std::map<XY, Block>::const_iterator it = blocks()->find(XY(x0 + loop_x, y0 + loop_y));
					int dst_x = (x0 + loop_x) * 64 - x;
					int dst_y = (y0 + loop_y) * 64 - y;
					if (it == blocks()->end()) {
						fill_pixel(dst, w, w, h, T(), dst_x, dst_y, w, h);
					} else {
						Block const *b = &it->second;
						b->_core->render(dst, w, w, h, dst_x, dst_y);
					}
				}
			}
		}

		void Image::render(int x, int y, int w, int h, PixelRGB *dst) const
		{
			renderT(x, y, w, h, dst);
		}

		void Image::render(int x, int y, int w, int h, PixelY *dst) const
		{
			renderT(x, y, w, h, dst);
		}

		void Image::invert(int x, int y, int w, int h)
		{
			switch (_core->_pixel_type) {
			case PT_RGBA:
				{
					std::vector<PixelRGB> tmp(w * h);
					render(x, y, w, h, &tmp[0]);
					size_t i, n;
					n = tmp.size();
					for (i = 0; i < n; i++) {
						PixelRGB *p = &tmp[i];
						if (p->a != 0) {
							p->r = ~p->r;
							p->g = ~p->g;
							p->b = ~p->b;
						}
					}
					put(x, y, &tmp[0], w, 0, 0, w, h);
				}
				break;
			case PT_YA:
				{
					std::vector<PixelY> tmp(w * h);
					render(x, y, w, h, &tmp[0]);
					size_t i, n;
					n = tmp.size();
					for (i = 0; i < n; i++) {
						PixelY *p = &tmp[i];
						if (p->a != 0) {
							p->y = ~p->y;
						}
					}
					put(x, y, &tmp[0], w, 0, 0, w, h);
				}
				break;
			}
		}

		template <typename PIXEL> static void wave_(int w, int h, double ox, double oy, double freq, double span, double angle, PIXEL const *src, PIXEL *dst)
		{
			double rad = angle * M_PI / 180;
			for (int y = 0; y < h; y++) {
				for (int x = 0; x < w; x++) {
					double px = x - ox;
					double py = y - oy;
					double t;
					t = fmod(sqrt(px * px + py * py), freq);
					t = span * sin(t * (M_PI * 2) / freq);
					px = x + sin(rad) * t + 0.5;
					py = y - cos(rad) * t + 0.5;
					*dst = get_pixel(src, w, h, px, py);
					dst++;
				}
			}
		}

		void Image::wave(int x, int y, int w, int h, double ox, double oy, double freq, double span, double angle)
		{
			switch (_core->_pixel_type) {
			case PT_RGBA:
				{
					std::vector<PixelRGB> src(w * h);
					std::vector<PixelRGB> dst(w * h);
					render(x, y, w, h, &src[0]);

					wave_(w, h, ox, oy, freq, span, angle, &src[0], &dst[0]);

					put(x, y, &dst[0], w, 0, 0, w, h);
				}
				break;
			case PT_YA:
				{
					std::vector<PixelY> src(w * h);
					std::vector<PixelY> dst(w * h);
					render(x, y, w, h, &src[0]);

					wave_(w, h, ox, oy, freq, span, angle, &src[0], &dst[0]);

					put(x, y, &dst[0], w, 0, 0, w, h);
				}
				break;
			}
		}

		template <typename PIXEL> static void rotate_(int w, int h, double angle, PIXEL const *src, PIXEL *dst)
		{
			memcpy(dst, src, sizeof(PIXEL) * w * h);
			if (angle == 0) {
				return;
			}
			double rad = angle * M_PI / 180;
			double s = sin(rad);
			double c = cos(rad);
			double ox = w / 2.0;
			double oy = h / 2.0;
			for (int y = 0; y < h; y++) {
				for (int x = 0; x < w; x++) {
					double x0 = x - ox + 0.5;
					double y0 = y - oy + 0.5;
					double x1 = x0 * c + y0 * s + ox;
					double y1 = y0 * c - x0 * s + oy;
					if (x1 < 0 || y1 < 0 || x1 > w || y1 > h) {
						dst->a = 0;
					} else {
						*dst = get_pixel(src, w, h, x1, y1);
					}
					dst++;
				}
			}
		}

		void rotate(int w, int h, double angle, PixelRGB const *src, PixelRGB *dst)
		{
			rotate_(w, h, angle, &src[0], &dst[0]);
		}

		void rotate(int w, int h, double angle, PixelY const *src, PixelY *dst)
		{
			rotate_(w, h, angle, &src[0], &dst[0]);
		}

		class median_t {
		private:
			int _map256[256];
			int _map16[16];
		public:
			median_t()
			{
				clear();
			}
			void clear()
			{
				for (int i = 0; i < 256; i++) {
					_map256[i] = 0;
				}
				for (int i = 0; i < 16; i++) {
					_map16[i] = 0;
				}
			}
			void insert(unsigned char n)
			{
				_map256[n]++;
				_map16[n >> 4]++;
			}
			void remove(unsigned char n)
			{
				_map256[n]--;
				_map16[n >> 4]--;
			}
			unsigned char get()
			{
				int left, right;
				int lower, upper;
				lower = 0;
				upper = 0;
				left = 0;
				right = 15;
				while (left < right) {
					if (lower + _map16[left] < upper + _map16[right]) {
						lower += _map16[left];
						left++;
					} else {
						upper += _map16[right];
						right--;
					}
				}
				left *= 16;
				right = left + 15;
				while (left < right) {
					if (lower + _map256[left] < upper + _map256[right]) {
						lower += _map256[left];
						left++;
					} else {
						upper += _map256[right];
						right--;
					}
				}
				return left;
			}

		};

		struct median_filter_rgb_t {
			median_t r;
			median_t g;
			median_t b;
			void insert(PixelRGB const &p)
			{
				r.insert(p.r);
				g.insert(p.g);
				b.insert(p.b);
			}
			void remove(PixelRGB const &p)
			{
				r.remove(p.r);
				g.remove(p.g);
				b.remove(p.b);
			}
			PixelRGB get(unsigned char a)
			{
				return PixelRGB(r.get(), g.get(), b.get(), a);
			}
		};

		struct median_filter_y_t {
			median_t y;
			void insert(PixelY const &p)
			{
				y.insert(p.y);
			}
			void remove(PixelY const &p)
			{
				y.remove(p.y);
			}
			PixelY get(unsigned char a)
			{
				return PixelY(y.get(), a);
			}
		};

		template <typename PIXEL, typename FILTER> void Image::medianT(int x, int y, int w, int h, int span)
		{
			int sw = w + span * 2;
			int sh = h + span * 2;
			std::vector<PIXEL> src(sw * sh);
			std::vector<PIXEL> dst(w * h);

			render(-span, -span, w + span * 2, h + span * 2, &src[0]);

			for (int y = 0; y < h; y++) {
				FILTER filter;
				for (int x = 0; x < span * 2; x++) {
					for (int i = 0; i < span * 2 + 1; i++) {
						PIXEL rgb = src[(y + i) * sw + x];
						if (rgb.a > 0) {
							filter.insert(rgb);
						}
					}
				}
				for (int x = 0; x < w; x++) {
					for (int i = 0; i < span * 2 + 1; i++) {
						PIXEL rgb = src[(y + i) * sw + span * 2 + x];
						if (rgb.a > 0) {
							filter.insert(rgb);
						}
					}
					PIXEL rgb = src[(span + y) * sw + span + x];
					if (rgb.a > 0) {
						rgb = filter.get(rgb.a);
					}

					dst[y * w + x] = rgb;

					for (int i = 0; i < span * 2 + 1; i++) {
						PIXEL rgb = src[(y + i) * sw + x];
						if (rgb.a > 0) {
							filter.remove(rgb);
						}
					}
				}
			}

			put(0, 0, &dst[0], w, 0, 0, w, h, BLEND_NORMAL, 255);
		}

		void Image::median(int x, int y, int w, int h, int span)
		{
			switch (_core->_pixel_type) {
			case PT_RGBA:
				medianT<PixelRGB, median_filter_rgb_t>(x, y, w, h, span);
				break;
			case PT_YA:
				medianT<PixelY, median_filter_y_t>(x, y, w, h, span);
				break;
			}
		}

		class minimize_t {
		private:
			int _map256[256];
			int _map16[16];
		public:
			minimize_t()
			{
				clear();
			}
			void clear()
			{
				for (int i = 0; i < 256; i++) {
					_map256[i] = 0;
				}
				for (int i = 0; i < 16; i++) {
					_map16[i] = 0;
				}
			}
			void insert(unsigned char n)
			{
				_map256[n]++;
				_map16[n >> 4]++;
			}
			void remove(unsigned char n)
			{
				_map256[n]--;
				_map16[n >> 4]--;
			}
			unsigned char get()
			{
				int left, right;
				for (left = 0; left < 16; left++) {
					if (_map16[left] != 0) {
						left *= 16;
						right = left + 16;
						while (left < right) {
							if (_map256[left] != 0) {
								return left;
							}
							left++;
						}
						break;
					}
				}
				return 0;
			}

		};

		struct minimize_filter_rgb_t {
			minimize_t r;
			minimize_t g;
			minimize_t b;
			void insert(PixelRGB const &p)
			{
				r.insert(p.r);
				g.insert(p.g);
				b.insert(p.b);
			}
			void remove(PixelRGB const &p)
			{
				r.remove(p.r);
				g.remove(p.g);
				b.remove(p.b);
			}
			PixelRGB get(unsigned char a)
			{
				return PixelRGB(r.get(), g.get(), b.get(), a);
			}
		};

		struct minimize_filter_y_t {
			minimize_t y;
			void insert(PixelY const &p)
			{
				y.insert(p.y);
			}
			void remove(PixelY const &p)
			{
				y.remove(p.y);
			}
			PixelY get(unsigned char a)
			{
				return PixelY(y.get(), a);
			}
		};

		template <typename PIXEL, typename FILTER> void Image::minimizeT(int x, int y, int w, int h, int span)
		{
			int sw = w + span * 2;
			int sh = h + span * 2;
			std::vector<PIXEL> src(sw * sh);
			std::vector<PIXEL> dst(w * h);

			render(-span, -span, w + span * 2, h + span * 2, &src[0]);

			for (int y = 0; y < h; y++) {
				FILTER filter;
				for (int x = 0; x < span * 2; x++) {
					for (int i = 0; i < span * 2 + 1; i++) {
						PIXEL rgb = src[(y + i) * sw + x];
						if (rgb.a > 0) {
							filter.insert(rgb);
						}
					}
				}
				for (int x = 0; x < w; x++) {
					for (int i = 0; i < span * 2 + 1; i++) {
						PIXEL rgb = src[(y + i) * sw + span * 2 + x];
						if (rgb.a > 0) {
							filter.insert(rgb);
						}
					}
					PIXEL rgb = src[(span + y) * sw + span + x];
					if (rgb.a > 0) {
						rgb = filter.get(rgb.a);
					}

					dst[y * w + x] = rgb;

					for (int i = 0; i < span * 2 + 1; i++) {
						PIXEL rgb = src[(y + i) * sw + x];
						if (rgb.a > 0) {
							filter.remove(rgb);
						}
					}
				}
			}

			put(0, 0, &dst[0], w, 0, 0, w, h, BLEND_NORMAL, 255);
		}

		void Image::minimize(int x, int y, int w, int h, int span)
		{
			switch (_core->_pixel_type) {
			case PT_RGBA:
				minimizeT<PixelRGB, minimize_filter_rgb_t>(x, y, w, h, span);
				break;
			case PT_YA:
				minimizeT<PixelY, minimize_filter_y_t>(x, y, w, h, span);
				break;
			}
		}

		class maximize_t {
		private:
			int _map256[256];
			int _map16[16];
		public:
			maximize_t()
			{
				clear();
			}
			void clear()
			{
				for (int i = 0; i < 256; i++) {
					_map256[i] = 0;
				}
				for (int i = 0; i < 16; i++) {
					_map16[i] = 0;
				}
			}
			void insert(unsigned char n)
			{
				_map256[n]++;
				_map16[n >> 4]++;
			}
			void remove(unsigned char n)
			{
				_map256[n]--;
				_map16[n >> 4]--;
			}
			unsigned char get()
			{
				int left, right;
				right = 16;
				while (right > 0) {
					right--;
					if (_map16[right] != 0) {
						left = right * 16;
						right = left + 16;
						while (left < right) {
							right--;
							if (_map256[right] != 0) {
								return right;
							}
						}
						break;
					}
				}
				return 0;
			}

		};

		struct maximize_filter_rgb_t {
			maximize_t r;
			maximize_t g;
			maximize_t b;
			void insert(PixelRGB const &p)
			{
				r.insert(p.r);
				g.insert(p.g);
				b.insert(p.b);
			}
			void remove(PixelRGB const &p)
			{
				r.remove(p.r);
				g.remove(p.g);
				b.remove(p.b);
			}
			PixelRGB get(unsigned char a)
			{
				return PixelRGB(r.get(), g.get(), b.get(), a);
			}
		};

		struct maximize_filter_y_t {
			maximize_t y;
			void insert(PixelY const &p)
			{
				y.insert(p.y);
			}
			void remove(PixelY const &p)
			{
				y.remove(p.y);
			}
			PixelY get(unsigned char a)
			{
				return PixelY(y.get(), a);
			}
		};

		template <typename PIXEL, typename FILTER> void Image::maximizeT(int x, int y, int w, int h, int span)
		{
			int sw = w + span * 2;
			int sh = h + span * 2;
			std::vector<PIXEL> src(sw * sh);
			std::vector<PIXEL> dst(w * h);

			render(-span, -span, w + span * 2, h + span * 2, &src[0]);

			for (int y = 0; y < h; y++) {
				FILTER filter;
				FILTER filter_g;
				FILTER filter_b;
				for (int x = 0; x < span * 2; x++) {
					for (int i = 0; i < span * 2 + 1; i++) {
						PIXEL rgb = src[(y + i) * sw + x];
						if (rgb.a > 0) {
							filter.insert(rgb);
						}
					}
				}
				for (int x = 0; x < w; x++) {
					for (int i = 0; i < span * 2 + 1; i++) {
						PIXEL rgb = src[(y + i) * sw + span * 2 + x];
						if (rgb.a > 0) {
							filter.insert(rgb);
						}
					}
					PIXEL rgb = src[(span + y) * sw + span + x];
					if (rgb.a > 0) {
						rgb = filter.get(rgb.a);
					}

					dst[y * w + x] = rgb;

					for (int i = 0; i < span * 2 + 1; i++) {
						PIXEL rgb = src[(y + i) * sw + x];
						if (rgb.a > 0) {
							filter.remove(rgb);
						}
					}
				}
			}

			put(0, 0, &dst[0], w, 0, 0, w, h, BLEND_NORMAL, 255);
		}

		void Image::maximize(int x, int y, int w, int h, int span)
		{
			switch (_core->_pixel_type) {
			case PT_RGBA:
				maximizeT<PixelRGB, maximize_filter_rgb_t>(x, y, w, h, span);
				break;
			case PT_YA:
				maximizeT<PixelRGB, maximize_filter_y_t>(x, y, w, h, span);
				break;
			}
		}

		struct FPixelRGB {
			double r, g, b, a, n;
			FPixelRGB()
				: r(0)
				, g(0)
				, b(0)
				, a(0)
				, n(0)
			{
			}
			FPixelRGB(double r, double g, double b, double a)
				: r(r)
				, g(g)
				, b(b)
				, a(a)
				, n(1)
			{
			}
			void add(PixelRGB const &t)
			{
				r += t.r * t.a;
				g += t.g * t.a;
				b += t.b * t.a;
				a += t.a;
				n++;
			}
			void sub(PixelRGB const &t)
			{
				r -= t.r * t.a;
				g -= t.g * t.a;
				b -= t.b * t.a;
				a -= t.a;
				n--;
			}
			operator PixelRGB ()
			{
				PixelRGB pixel;
				pixel.r = (unsigned char)(r / a);
				pixel.g = (unsigned char)(g / a);
				pixel.b = (unsigned char)(b / a);
				pixel.a = (unsigned char)(a / n);
				return pixel;
			}
		};

		struct FPixelY {
			double y, a, n;
			FPixelY()
				: y(0)
				, a(0)
				, n(0)
			{
			}
			FPixelY(double y, double a)
				: y(y)
				, a(a)
				, n(1)
			{
			}
			void add(PixelY const &t)
			{
				y += t.y * t.a;
				a += t.a;
				n++;
			}
			void sub(PixelY const &t)
			{
				y -= t.y * t.a;
				a -= t.a;
				n--;
			}
			operator PixelY ()
			{
				PixelY pixel;
				pixel.y = (unsigned char)(y / a);
				pixel.a = (unsigned char)(a / n);
				return pixel;
			}
		};

		template <typename T, typename F> void Image::blurT(T *src, T *dst, int n, int span, int loop)
		{
			for (int pass = 0; pass < loop; pass++) {
				if (pass > 0) {
					std::copy(dst, dst + n, src);
				}

				F fpix;

				for (int i = 0; i < span; i++) {
					fpix.add(src[i]);
				}

				for (int i = 0; i < n; i++) {
					if (span + i < n) {
						fpix.add(src[span + i]);
					}

					dst[i] = (T)fpix;

					if (i >= span) {
						fpix.sub(src[i - span]);
					}
				}
			}
		}

		template <typename T, typename F> void Image::blur_h(int x, int y, int w, int span, int loop)
		{
			std::vector<T> src(w);
			std::vector<T> dst(w);
			render(x, y, w, 1, &src[0]);
			blurT<T, F>(&src[0], &dst[0], w, span, loop);
			put(x, y, &dst[0], w, 0, 0, w, 1, BLEND_NORMAL, 255);
		}

		template <typename T, typename F> void Image::blur_v(int x, int y, int h, int span, int loop)
		{
			std::vector<T> src(h);
			std::vector<T> dst(h);
			render(x, y, 1, h, &src[0]);
			blurT<T, F>(&src[0], &dst[0], h, span, loop);
			put(x, y, &dst[0], 1, 0, 0, 1, h, BLEND_NORMAL, 255);
		}

		void Image::blur(int x, int y, int w, int h, int span_h, int span_v, int loop)
		{
			switch (_core->_pixel_type) {
			case PT_RGBA:
				for (int i = 0; i < h; i++) {
					blur_h<PixelRGB, FPixelRGB>(x, y + i, w, span_h, loop);
				}
				for (int i = 0; i < w; i++) {
					blur_v<PixelRGB, FPixelRGB>(x + i, y, h, span_v, loop);
				}
				break;
			case PT_YA:
				for (int i = 0; i < h; i++) {
					blur_h<PixelY, FPixelY>(x, y + i, w, span_h, loop);
				}
				for (int i = 0; i < w; i++) {
					blur_v<PixelY, FPixelY>(x + i, y, h, span_v, loop);
				}
				break;
			}
		}

		void Image::zebra(int x, int y, int w, int h, int span, int loop)
		{
			int i, n;
			n = w * h;
			switch (_core->_pixel_type) {
			case PT_RGBA:
				for (int pass = 0; pass < loop; pass++) {
					std::vector<PixelRGB> foo(n);
					std::vector<PixelRGB> bar(n);
					render(0, 0, w, h, &foo[0]);
					maximize(x, y, w, h, span);
					render(0, 0, w, h, &bar[0]);
					for (i = 0; i < n; i++) {
						PixelRGB *src = &foo[i];
						PixelRGB *dst = &bar[i];
						if (src->a != 0 && dst->a != 0) {
							dst->r = (dst->r * (255 - src->r) * 257 + 32768) / 65536;
							dst->g = (dst->g * (255 - src->g) * 257 + 32768) / 65536;
							dst->b = (dst->b * (255 - src->b) * 257 + 32768) / 65536;
						}
					}
					put(x, y, &bar[0], w, 0, 0, w, h);
				}
				break;
			case PT_YA:
				for (int pass = 0; pass < loop; pass++) {
					std::vector<PixelY> foo(n);
					std::vector<PixelY> bar(n);
					render(0, 0, w, h, &foo[0]);
					maximize(x, y, w, h, span);
					render(0, 0, w, h, &bar[0]);
					for (i = 0; i < n; i++) {
						PixelY *src = &foo[i];
						PixelY *dst = &bar[i];
						if (src->a != 0 && dst->a != 0) {
							dst->y = (dst->y * (255 - src->y) * 257 + 32768) / 65536;
						}
					}
					put(x, y, &bar[0], w, 0, 0, w, h);
				}
				break;
			}
		}

		///////////////////////////////////////////////////////////////////////////////

		void copy(Image *dst, int dst_x, int dst_y, Image const *src, int src_x, int src_y, int w, int h, BlendMode blend, int opacity)
		{
			std::vector<PixelRGB> tmp(64 * 64);

			int x, y;

			for (y = 0; y < h; y += 64) {
				for (x = 0; x < w; x += 64) {
					int i = w - x;
					int j = h - y;
					if (i > 64) i = 64;
					if (j > 64) j = 64;
					src->render(src_x + x, src_y + y, i, j, &tmp[0]);
					dst->put(dst_x + x, dst_y + y, &tmp[0], i, 0, 0, i, j, blend, opacity);
				}
			}
		}

		///////////////////////////////////////////////////////////////////////////////

		///////////////////////////////////////////////////////////////////////////////
	}
}


