
#ifndef __LUMINOUS_H
#define __LUMINOUS_H

#include <map>
#include <vector>

namespace soramimi {
	namespace image {

		struct PixelRGB;
		class BlockCoreRGB;
		struct PixelY;
		class BlockCoreY;

		enum pixel_type_t {
			PT_RGBA,
			PT_YA,
		};

		struct PixelRGB {
			typedef BlockCoreRGB BlockType;
			unsigned char r, g, b, a;
			PixelRGB()
				: r(0)
				, g(0)
				, b(0)
				, a(0)
			{
			}
			PixelRGB(unsigned char r, unsigned char g, unsigned char b, unsigned char a)
				: r(r)
				, g(g)
				, b(b)
				, a(a)
			{
			}
			PixelRGB(PixelY const &pixel);
		};

		struct PixelY {
			typedef BlockCoreY BlockType;
			unsigned char y, a;
			PixelY()
				: y(0)
				, a(0)
			{
			}
			PixelY(unsigned char y, unsigned char a)
				: y(y)
				, a(a)
			{
			}
			PixelY(PixelRGB const &pixel);
		};

		inline unsigned char gray(unsigned char r, unsigned char g, unsigned char b)
		{
			return (306 * r + 601 * g + 117 * b + 512) / 1024;
		}

		inline unsigned char gray(PixelRGB const &rgba)
		{
			return rgba.a == 0 ? 0 : gray(rgba.r, rgba.g, rgba.b);
		}

		inline PixelRGB::PixelRGB(PixelY const &pixel)
			: r(pixel.y)
			, g(pixel.y)
			, b(pixel.y)
			, a(pixel.a)
		{
		}

		inline PixelY::PixelY(PixelRGB const &pixel)
			: y(gray(pixel))
			, a(pixel.a)
		{
		}

		enum BlendMode {
			BLEND_NORMAL,
			BLEND_XOR,
			BLEND_MIN,
			BLEND_MAX,
		};

		class BlockCore {
			friend class Block;
		private:
			unsigned int _reference_count;
		protected:
			virtual BlockCore *clone() const = 0;
			virtual BlockCore *convert(pixel_type_t pt) const = 0;
		public:
			virtual ~BlockCore()
			{
			}
			BlockCore()
				: _reference_count(0)
			{
			}
			virtual void fill(int x, int y, int w, int h, PixelRGB const &src) = 0;
			virtual void put(int x, int y, PixelRGB const *src, int src_span, int src_x, int src_y, int w, int h, BlendMode blend, int opacity) = 0;
			virtual void render(PixelRGB *dst, int dst_span, int dst_w, int dst_h, int dst_x, int dst_y) const = 0;
			virtual void fill(int x, int y, int w, int h, PixelY const &src) = 0;
			virtual void put(int x, int y, PixelY const *src, int src_span, int src_x, int src_y, int w, int h, BlendMode blend, int opacity) = 0;
			virtual void render(PixelY *dst, int dst_span, int dst_w, int dst_h, int dst_x, int dst_y) const = 0;
		};

		class Block {
			friend class Image;
			friend class BlockCoreRGB;
			friend class BlockCoreY;
		private:
			int _ox;
			int _oy;
			BlockCore *_core;
			void assign(BlockCore *p)
			{
				if (p) {
					p->_reference_count++;
				}
				if (_core) {
					if (_core->_reference_count > 1) {
						_core->_reference_count--;
					} else {
						delete _core;
					}
				}
				_core = p;
			}
		private:
			Block(BlockCore *core)
				: _core(0)
				, _ox(0)
				, _oy(0)
			{
				assign(core);
			}
		public:
			~Block()
			{
				assign(0);
			}
			Block()
				: _core(0)
				, _ox(0)
				, _oy(0)
			{
			}
			Block(Block const &r)
				: _core(0)
				, _ox(r._ox)
				, _oy(r._oy)
			{
				assign(r._core);
			}
			void operator = (Block const &r)
			{
				assign(r._core);
				_ox = r._ox;
				_oy = r._oy;
			}
			operator bool () const
			{
				return _core != 0;
			}
		private:
			void copy_to(Block *to)
			{
				BlockCore *b = 0;
				if (_core) {
					b = _core->clone();
				}
				to->assign(b);
			}
		public:
			void independent()
			{
				if (_core && _core->_reference_count > 1) {
					Block b;
					copy_to(&b);
					assign(b._core);
				}
			}
			void convert(pixel_type_t pt);
		};

		class BlockCoreRGB : public BlockCore {
			friend class BlockCoreY;
		public:
			typedef PixelRGB PixelType;
		private:
			void clear();
		protected:
			PixelRGB _data[64][64];
			virtual BlockCore *clone() const;
			virtual BlockCore *convert(pixel_type_t pt) const;
		public:
			BlockCoreRGB()
			{
				clear();
			}
			static Block NEW()
			{
				return Block(new BlockCoreRGB());
			}
			virtual void fill(int x, int y, int w, int h, PixelRGB const &src);
			virtual void put(int x, int y, PixelRGB const *src, int src_span, int src_x, int src_y, int w, int h, BlendMode blend, int opacity);
			virtual void render(PixelRGB *dst, int dst_span, int dst_w, int dst_h, int dst_x, int dst_y) const;
			virtual void fill(int x, int y, int w, int h, PixelY const &src);
			virtual void put(int x, int y, PixelY const *src, int src_span, int src_x, int src_y, int w, int h, BlendMode blend, int opacity);
			virtual void render(PixelY *dst, int dst_span, int dst_w, int dst_h, int dst_x, int dst_y) const;
		};

		class BlockCoreY : public BlockCore {
			friend class BlockCoreRGB;
		public:
			typedef PixelY PixelType;
		private:
			void clear();
		protected:
			PixelY _data[64][64];
			virtual BlockCore *clone() const;
			virtual BlockCore *convert(pixel_type_t pt) const;
		public:
			BlockCoreY()
			{
				clear();
			}
			static Block NEW()
			{
				return Block(new BlockCoreY());
			}
			virtual void fill(int x, int y, int w, int h, PixelRGB const &src);
			virtual void put(int x, int y, PixelRGB const *src, int src_span, int src_x, int src_y, int w, int h, BlendMode blend, int opacity);
			virtual void render(PixelRGB *dst, int dst_span, int dst_w, int dst_h, int dst_x, int dst_y) const;
			virtual void fill(int x, int y, int w, int h, PixelY const &src);
			virtual void put(int x, int y, PixelY const *src, int src_span, int src_x, int src_y, int w, int h, BlendMode blend, int opacity);
			virtual void render(PixelY *dst, int dst_span, int dst_w, int dst_h, int dst_x, int dst_y) const;
		};

		struct XY {
			int x;
			int y;
			XY(int x, int y)
				: x(x)
				, y(y)
			{
			}
		};

		static inline bool operator < (XY const &a, XY const &b)
		{
			return (a.y < b.y) || (a.y == b.y && a.x < b.x);
		}

		class Image {
		public:
		private:
			struct ImageCore {
				unsigned int _reference_count;
				std::map<XY, Block> _blocks;
				pixel_type_t _pixel_type;
				ImageCore(pixel_type_t pt)
					: _reference_count(0)
					, _pixel_type(pt)
				{
				}
			};
			ImageCore *_core;
			void assign(ImageCore *p);

			std::map<XY, Block> const *blocks() const
			{
				return &_core->_blocks;
			}
			std::map<XY, Block> *blocks()
			{
				return &_core->_blocks;
			}

		public:
			Image();
			~Image();
			Image(Image const &r)
				: _core(0)
			{
				assign(r._core);
			}
			void operator = (Image const &r)
			{
				assign(r._core);
			}
			void clear(pixel_type_t pt);
		private:
			void independent();
		public:
			void convert(pixel_type_t pt);
		private:
			Block newblock() const;
			template <typename T> void fillT(int dst_x, int dst_y, T const &src, int src_span, int src_x, int src_y, int src_w, int src_h);
			template <typename T> void fillT(int x, int y, int w, int h, T const &src);
			template <typename T> void putT(int dst_x, int dst_y, T *src, int src_span, int src_x, int src_y, int src_w, int src_h, BlendMode blend, int opacity);
			template <typename T> void renderT(int x, int y, int w, int h, T *dst) const;
		public:
			void fill(int x, int y, int w, int h, PixelRGB const &src);
			void fill(int x, int y, int w, int h, PixelY const &src);
			void put(int dst_x, int dst_y, PixelRGB *src, int src_span, int src_x, int src_y, int src_w, int src_h, BlendMode blend = BLEND_NORMAL, int opacity = 255);
			void put(int dst_x, int dst_y, PixelY *src, int src_span, int src_x, int src_y, int src_w, int src_h, BlendMode blend = BLEND_NORMAL, int opacity = 255);
			void render(int x, int y, int w, int h, PixelRGB *dst) const;
			void render(int x, int y, int w, int h, PixelY *dst) const;
		public:
			void invert(int x, int y, int w, int h);
			void wave(int x, int y, int w, int h, double ox, double oy, double freq, double span, double angle);
		private:
			template <typename PIXEL, typename FILTER> void medianT(int x, int y, int w, int h, int span);
			template <typename PIXEL, typename FILTER> void minimizeT(int x, int y, int w, int h, int span);
			template <typename PIXEL, typename FILTER> void maximizeT(int x, int y, int w, int h, int span);
		public:
			void median(int x, int y, int w, int h, int span);
			void minimize(int x, int y, int w, int h, int span);
			void maximize(int x, int y, int w, int h, int span);
		private:
			template <typename T, typename F> void blurT(T *src, T *dst, int n, int span, int loop);
			template <typename T, typename F> void blur_h(int x, int y, int w, int span, int loop);
			template <typename T, typename F> void blur_v(int x, int y, int h, int span, int loop);
		public:
			void blur(int x, int y, int w, int h, int span_h, int span_v, int loop);
		public:
			void zebra(int x, int y, int w, int h, int span, int loop);
		};

		void rotate(int w, int h, double angle, PixelRGB const *src, PixelRGB *dst);
		void rotate(int w, int h, double angle, PixelY const *src, PixelY *dst);

		void copy(Image *dst, int dst_x, int dst_y, Image const *src, int src_x, int src_y, int w, int h, BlendMode blend = BLEND_NORMAL, int opacity = 255);

		bool write_jpeg(std::vector<unsigned char> *vec, soramimi::image::Image const *src, int width, int height, int quality = 75);
		bool write_jpeg(char const *path, soramimi::image::Image const *src, int width, int height, int quality);
		bool write_png(char const *filename, soramimi::image::Image const *src, int width, int height);
		bool read_jpeg(soramimi::image::Image *dst, char const *filename, int *width_ptr, int *height_ptr);
		bool read_png(soramimi::image::Image *dst, char const *filename, int *width_p, int *height_p);
		bool write_png(std::vector<unsigned char> *vec, soramimi::image::Image const *src, int width, int height);
	}
}





#endif

