

#include <png.h>
#include <vector>
#include <stdio.h>

#include "image/image.h"

#pragma warning(disable:4996)

namespace soramimi {
	namespace image {

		static void png_cexcept_error(png_structp png_ptr, png_const_charp msg)
		{
			 throw msg;
		}

		static bool write_png(soramimi::image::Image const *src, int width, int height, void (*method)(png_structp png_ptr, png_bytep data, png_size_t length), void *cookie)
		{
			bool success = false;

			png_structp png_ptr = NULL;
			png_infop info_ptr = NULL;

			const int           ciBitDepth = 8;
			const int           ciChannels = 3;

			static FILE        *pfFile;
			png_uint_32         ulRowBytes;

			// prepare the standard PNG structures

			png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, (png_error_ptr)png_cexcept_error, (png_error_ptr)NULL);
			if (!png_ptr) {
				return success;
			}

			info_ptr = png_create_info_struct(png_ptr);
			if (!info_ptr) {
				png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
				return success;
			}

			try {
				// initialize the png structure

				png_set_write_fn(png_ptr, cookie, method, NULL);

				// we're going to write a very simple 3x8 bit RGB image

				png_set_IHDR(png_ptr, info_ptr, width, height, ciBitDepth, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

				// write the file header information

				png_write_info(png_ptr, info_ptr);

				// swap the BGR pixels in the DiData structure to RGB

				png_set_bgr(png_ptr);

				// row_bytes is the width x number of channels

				ulRowBytes = width * ciChannels;

				std::vector<soramimi::image::PixelRGB> tmp_rgba(width);
				std::vector<unsigned char> tmp_rgb(ulRowBytes);

				{
					int i;
					int pass, num_pass; /* pass variables */

					png_debug(1, "in png_write_image\n");
					/* intialize interlace handling.  If image is not interlaced,
					this will set pass to 1 */
					num_pass = png_set_interlace_handling(png_ptr);

					/* loop through passes */
					for (pass = 0; pass < num_pass; pass++) {
						/* loop through image */
						for (i = 0; i < height; i++) {
							src->render(0, i, width, 1, &tmp_rgba[0]);
							for (int i = 0; i < width; i++) {
								tmp_rgb[i * 3 + 0] = tmp_rgba[i].b;
								tmp_rgb[i * 3 + 1] = tmp_rgba[i].g;
								tmp_rgb[i * 3 + 2] = tmp_rgba[i].r;
							}
							png_write_row(png_ptr, &tmp_rgb[0]);
						}
					}
				}

				// write the additional chunks to the PNG file (not really needed)

				png_write_end(png_ptr, info_ptr);

				success = true;

			} catch (png_const_charp) {
			}

			png_destroy_write_struct(&png_ptr, &info_ptr);

			return success;
		}

		static void file_writer(png_structp png_ptr, png_bytep data, png_size_t length)
		{
			void *cookie = png_ptr->io_ptr;
			fwrite(data, 1, length, (FILE *)cookie);
		}

		static void vector_writer(png_structp png_ptr, png_bytep data, png_size_t length)
		{
			void *cookie = png_ptr->io_ptr;
			std::vector<unsigned char> *vec = (std::vector<unsigned char> *)cookie;
			vec->insert(vec->end(), data, data + length);
		}


		static bool write_png(FILE *file, soramimi::image::Image const *src, int width, int height)
		{
			return write_png(src, width, height, file_writer, file);
		}

		bool write_png(char const *filename, soramimi::image::Image const *src, int width, int height)
		{
			FILE *fp;
			fp = fopen(filename, "wb");
			if (!fp) {
				return false;
			}

			bool success = write_png(fp, src, width, height);

			fclose(fp);

			return success;
		}

		bool write_png(std::vector<unsigned char> *vec, soramimi::image::Image const *src, int width, int height)
		{
			return write_png(src, width, height, vector_writer, vec);
		}



		typedef unsigned char byte;


		struct Source {
			int (*method)(void *ptr, size_t len, void *cookie);
			void *cookie;
		};

		static void _get_bytes(png_structp png_ptr, png_bytep data, png_size_t length)
		{
			Source *info = (Source *)png_ptr->io_ptr;
			info->method(data, length, info->cookie);
		}

		static bool read_png(soramimi::image::Image *dst, int (*method)(void *ptr, size_t len, void *cookie), void *cookie, int *width_p, int *height_p)
		{
			bool success = false;

			png_structp png_ptr = NULL;
			png_infop info_ptr = NULL;
			png_byte            pbSig[8];
			int                 iBitDepth;
			int                 iColorType;
			double              dGamma;
			png_color_16       *pBackground;
			png_uint_32         ulChannels;
			png_uint_32         ulRowBytes;
			png_byte *rowbuffer = NULL;
			unsigned long width;
			unsigned long height;
			unsigned long bgcolor;
			Source source;

			source.method = method;
			source.cookie = cookie;

			// first check the eight byte PNG signature

			if (method(pbSig, 8, cookie) != 8) {
				return success;
			}
			if (!png_check_sig(pbSig, 8)) {
				return success;
			}

			// create the two png(-info) structures

			png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, (png_error_ptr)png_cexcept_error, (png_error_ptr)NULL);
			if (!png_ptr) {
				return success;
			}

			info_ptr = png_create_info_struct(png_ptr);
			if (!info_ptr) {
				png_destroy_read_struct(&png_ptr, NULL, NULL);
				return success;
			}

			try {

				// initialize the png structure

				png_set_read_fn(png_ptr, (png_voidp)&source, _get_bytes);
				png_set_sig_bytes(png_ptr, 8);

				// read all PNG info up to image data

				png_read_info(png_ptr, info_ptr);

				// get width, height, bit-depth and color-type

				png_get_IHDR(png_ptr, info_ptr, &width, &height, &iBitDepth, &iColorType, NULL, NULL, NULL);

				// expand images of all color-type and bit-depth to 3x8 bit RGB images
				// let the library process things like alpha, transparency, background

				if (iBitDepth == 16)
					png_set_strip_16(png_ptr);

				if (iColorType == PNG_COLOR_TYPE_PALETTE)
					png_set_expand(png_ptr);

				if (iBitDepth < 8)
					png_set_expand(png_ptr);

				if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
					png_set_expand(png_ptr);

				if (iColorType == PNG_COLOR_TYPE_GRAY || iColorType == PNG_COLOR_TYPE_GRAY_ALPHA)
					png_set_gray_to_rgb(png_ptr);

				// set the background color to draw transparent and alpha images over.
				if (png_get_bKGD(png_ptr, info_ptr, &pBackground)) {
					png_set_background(png_ptr, pBackground, PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
					int r = (byte) pBackground->red;
					int g = (byte) pBackground->green;
					int b = (byte) pBackground->blue;
					bgcolor = 0xff000000 | (b << 16) | (g << 8) | b;
				} else {
					bgcolor = 0;
				}

				// if required set gamma conversion
				if (png_get_gAMA(png_ptr, info_ptr, &dGamma))
					png_set_gamma(png_ptr, (double) 2.2, dGamma);

				// after the transformations have been registered update info_ptr data

				png_read_update_info(png_ptr, info_ptr);

				// get again width, height and the new bit-depth and color-type

				png_get_IHDR(png_ptr, info_ptr, &width, &height, &iBitDepth, &iColorType, NULL, NULL, NULL);


				// row_bytes is the width x number of channels

				ulRowBytes = png_get_rowbytes(png_ptr, info_ptr);
				ulChannels = png_get_channels(png_ptr, info_ptr);

				// now we can allocate memory to store the image

				//dst->CreateImage(width, height, ulChannels, bgcolor);

				std::vector<png_byte> rowbuffer(ulRowBytes);
				std::vector<soramimi::image::PixelRGB> tmp_rgba(width);

				{
					png_uint_32 i,image_height;
					int pass, j;

					png_debug(1, "in png_read_image\n");
					/* save jump buffer and error functions */

					pass = png_set_interlace_handling(png_ptr);

					image_height=png_ptr->height;
					png_ptr->num_rows = image_height; /* Make sure this is set correctly */

					for (j = 0; j < pass; j++) {
						for (i = 0; i < image_height; i++) {
							png_read_row(png_ptr, &rowbuffer[0], NULL);
							if (ulChannels == 3) {
								png_byte *srcptr = &rowbuffer[0];
								soramimi::image::PixelRGB *dstptr = &tmp_rgba[0];
								for (unsigned int x = 0; x < width; x++) {
									dstptr->r = srcptr[0];
									dstptr->g = srcptr[1];
									dstptr->b = srcptr[2];
									dstptr->a = 255;
									srcptr += 3;
									dstptr++;
								}
								dst->put(0, i, &tmp_rgba[0], width, 0, 0, width, 1);
							} else if (ulChannels == 4) {
								png_byte *srcptr = &rowbuffer[0];
								soramimi::image::PixelRGB *dstptr = &tmp_rgba[0];
								for (unsigned int x = 0; x < width; x++) {
									dstptr->r = srcptr[0];
									dstptr->g = srcptr[1];
									dstptr->b = srcptr[2];
									dstptr->a = srcptr[3];
									srcptr += 4;
									dstptr++;
								}
								dst->put(0, i, &tmp_rgba[0], width, 0, 0, width, 1);
							}
						}
					}
				}

				// read the additional chunks in the PNG file (not really needed)

				png_read_end(png_ptr, NULL);

				success = true;		// success

			} catch (png_const_charp) {
			}

			png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
			//delete[] rowbuffer;

			*width_p = width;
			*height_p = height;

			return success;
		}


		static int myread(void *ptr, size_t len, void *cookie)
		{
			return (int)fread(ptr, 1, len, (FILE *)cookie);
		}


		bool read_png(soramimi::image::Image *dst, char const *filename, int *width_p, int *height_p)
		{
			FILE *fp = fopen(filename, "rb");
			if (!fp) {
				return false;
			}

			bool success = read_png(dst, myread, fp, width_p, height_p);

			fclose(fp);

			return success;
		}

	}
}
